import pika
import logging

import time

log = logging.getLogger("seabird_ctd")


class RabbitConsumer(object):
	"""
		Refactored the example consumer of messages in Pika's documentation to serve here.
	"""
	EXCHANGE = 'seabird'
	EXCHANGE_TYPE = 'topic'
	QUEUE = 'seabird'  # this should ultimately be set by the user, probably the calling function
	ROUTING_KEY = 'seabird'

	def __init__(self, amqp_url):
		"""Create a new instance of the consumer class, passing in the AMQP
		URL used to connect to RabbitMQ.

		:param str amqp_url: The AMQP url to connect with

		"""
		self._connection = None
		self._channel = None
		self._closing = False
		self._consumer_tag = None
		self.base_url = amqp_url

		self.username = None
		self.password = None
		self.vhost = None

		self.handler = None

	def connect(self):
		"""This method connects to RabbitMQ, returning the connection handle.
		When the connection is established, the on_connection_open method
		will be invoked by pika.

		:rtype: pika.SelectConnection

		"""
		self._url = "amqp://{}:{}@{}/{}".format(self.username, self.password, self.base_url, self.vhost)
		log.debug('Connecting to %s', self.base_url)
		return pika.SelectConnection(pika.URLParameters(self._url),
									 self.on_connection_open,
									 stop_ioloop_on_close=False)

	def on_connection_open(self, unused_connection):
		"""This method is called by pika once the connection to RabbitMQ has
		been established. It passes the handle to the connection object in
		case we need it, but in this case, we'll just mark it unused.

		:type unused_connection: pika.SelectConnection

		"""
		log.debug('Connection opened')
		self.add_on_connection_close_callback()
		self.open_channel()

	def add_on_connection_close_callback(self):
		"""This method adds an on close callback that will be invoked by pika
		when RabbitMQ closes the connection to the publisher unexpectedly.

		"""
		log.debug('Adding connection close callback')
		self._connection.add_on_close_callback(self.on_connection_closed)

	def on_connection_closed(self, connection, reply_code, reply_text):
		"""This method is invoked by pika when the connection to RabbitMQ is
		closed unexpectedly. Since it is unexpected, we will reconnect to
		RabbitMQ if it disconnects.

		:param pika.connection.Connection connection: The closed connection obj
		:param int reply_code: The server provided reply_code if given
		:param str reply_text: The server provided reply_text if given

		"""
		self._channel = None
		if self._closing:
			self._connection.ioloop.stop()
		else:
			log.warning('Connection closed, reopening in 5 seconds: (%s) %s',
						reply_code, reply_text)
			self._connection.add_timeout(5, self.reconnect)

	def reconnect(self):
		"""Will be invoked by the IOLoop timer if the connection is
		closed. See the on_connection_closed method.

		"""
		# This is the old connection IOLoop instance, stop its ioloop
		self._connection.ioloop.stop()

		if not self._closing:
			# Create a new connection
			self._connection = self.connect()

			# There is now a new connection, needs a new ioloop to run
			self._connection.ioloop.start()

	def open_channel(self):
		"""Open a new channel with RabbitMQ by issuing the Channel.Open RPC
		command. When RabbitMQ responds that the channel is open, the
		on_channel_open callback will be invoked by pika.

		"""
		log.debug('Creating a new channel')
		self._connection.channel(on_open_callback=self.on_channel_open)

	def on_channel_open(self, channel):
		"""This method is invoked by pika when the channel has been opened.
		The channel object is passed in so we can make use of it.

		Since the channel is now open, we'll declare the exchange to use.

		:param pika.channel.Channel channel: The channel object

		"""
		log.debug('Channel opened')
		self._channel = channel
		self.add_on_channel_close_callback()
		self.setup_exchange(self.EXCHANGE)

	def add_on_channel_close_callback(self):
		"""This method tells pika to call the on_channel_closed method if
		RabbitMQ unexpectedly closes the channel.

		"""
		log.debug('Adding channel close callback')
		self._channel.add_on_close_callback(self.on_channel_closed)

	def on_channel_closed(self, channel, reply_code, reply_text):
		"""Invoked by pika when RabbitMQ unexpectedly closes the channel.
		Channels are usually closed if you attempt to do something that
		violates the protocol, such as re-declare an exchange or queue with
		different parameters. In this case, we'll close the connection
		to shutdown the object.

		:param pika.channel.Channel: The closed channel
		:param int reply_code: The numeric reason the channel was closed
		:param str reply_text: The text reason the channel was closed

		"""
		log.warning('Channel %i was closed: (%s) %s',
					channel, reply_code, reply_text)
		self._connection.close()

	def setup_exchange(self, exchange_name):
		"""Setup the exchange on RabbitMQ by invoking the Exchange.Declare RPC
		command. When it is complete, the on_exchange_declareok method will
		be invoked by pika.

		:param str|unicode exchange_name: The name of the exchange to declare

		"""
		log.debug('Declaring exchange %s', exchange_name)
		self._channel.exchange_declare(self.on_exchange_declareok,
									   exchange_name,
									   self.EXCHANGE_TYPE)

	def on_exchange_declareok(self, unused_frame):
		"""Invoked by pika when RabbitMQ has finished the Exchange.Declare RPC
		command.

		:param pika.Frame.Method unused_frame: Exchange.DeclareOk response frame

		"""
		log.debug('Exchange declared')
		self.setup_queue(self.QUEUE)

	def setup_queue(self, queue_name):
		"""Setup the queue on RabbitMQ by invoking the Queue.Declare RPC
		command. When it is complete, the on_queue_declareok method will
		be invoked by pika.

		:param str|unicode queue_name: The name of the queue to declare.

		"""
		log.debug('Declaring queue %s', queue_name)
		self._channel.queue_declare(self.on_queue_declareok, queue_name)

	def on_queue_declareok(self, method_frame):
		"""Method invoked by pika when the Queue.Declare RPC call made in
		setup_queue has completed. In this method we will bind the queue
		and exchange together with the routing key by issuing the Queue.Bind
		RPC command. When this command is complete, the on_bindok method will
		be invoked by pika.

		:param pika.frame.Method method_frame: The Queue.DeclareOk frame

		"""
		log.debug('Binding %s to %s with %s',
				 self.EXCHANGE, self.QUEUE, self.ROUTING_KEY)
		self._channel.queue_bind(self.on_bindok, self.QUEUE,
								 self.EXCHANGE, self.ROUTING_KEY)

	def on_bindok(self, unused_frame):
		"""Invoked by pika when the Queue.Bind method has completed. At this
		point we will start consuming messages by calling start_consuming
		which will invoke the needed RPC commands to start the process.

		:param pika.frame.Method unused_frame: The Queue.BindOk response frame

		"""
		log.debug('Queue bound')
		self.start_consuming()

	def start_consuming(self):
		"""This method sets up the consumer by first calling
		add_on_cancel_callback so that the object is notified if RabbitMQ
		cancels the consumer. It then issues the Basic.Consume RPC command
		which returns the consumer tag that is used to uniquely identify the
		consumer with RabbitMQ. We keep the value to use it when we want to
		cancel consuming. The on_message method is passed in as a callback pika
		will invoke when a message is fully received.

		"""
		log.debug('Issuing consumer related RPC commands')
		self.add_on_cancel_callback()
		self._consumer_tag = self._channel.basic_consume(self.handler, self.QUEUE)

	def add_on_cancel_callback(self):
		"""Add a callback that will be invoked if RabbitMQ cancels the consumer
		for some reason. If RabbitMQ does cancel the consumer,
		on_consumer_cancelled will be invoked by pika.

		"""
		log.debug('Adding consumer cancellation callback')
		self._channel.add_on_cancel_callback(self.on_consumer_cancelled)

	def on_consumer_cancelled(self, method_frame):
		"""Invoked by pika when RabbitMQ sends a Basic.Cancel for a consumer
		receiving messages.

		:param pika.frame.Method method_frame: The Basic.Cancel frame

		"""
		log.debug('Consumer was cancelled remotely, shutting down: %r',
				 method_frame)
		if self._channel:
			self._channel.close()

	def on_message(self, unused_channel, basic_deliver, properties, body):
		"""Invoked by pika when a message is delivered from RabbitMQ. The
		channel is passed for your convenience. The basic_deliver object that
		is passed in carries the exchange, routing key, delivery tag and
		a redelivered flag for the message. The properties passed in is an
		instance of BasicProperties with the message properties and the body
		is the message that was sent.

		:param pika.channel.Channel unused_channel: The channel object
		:param pika.Spec.Basic.Deliver: basic_deliver method
		:param pika.Spec.BasicProperties: properties
		:param str|unicode body: The message body

		"""
		log.debug('Received message # %s from %s: %s', basic_deliver.delivery_tag, properties.app_id, body)
		self.acknowledge_message(basic_deliver.delivery_tag)

	def acknowledge_message(self, delivery_tag):
		"""Acknowledge the message delivery from RabbitMQ by sending a
		Basic.Ack RPC method for the delivery tag.

		:param int delivery_tag: The delivery tag from the Basic.Deliver frame

		"""
		log.debug('Acknowledging message %s', delivery_tag)
		self._channel.basic_ack(delivery_tag)

	def stop_consuming(self):
		"""Tell RabbitMQ that you would like to stop consuming by sending the
		Basic.Cancel RPC command.

		"""
		if self._channel:
			log.debug('Sending a Basic.Cancel RPC command to RabbitMQ')
			self._channel.basic_cancel(self.on_cancelok, self._consumer_tag)

	def on_cancelok(self, unused_frame):
		"""This method is invoked by pika when RabbitMQ acknowledges the
		cancellation of a consumer. At this point we will close the channel.
		This will invoke the on_channel_closed method once the channel has been
		closed, which will in-turn close the connection.

		:param pika.frame.Method unused_frame: The Basic.CancelOk frame

		"""
		log.debug('RabbitMQ acknowledged the cancellation of the consumer')
		self.close_channel()

	def close_channel(self):
		"""Call to close the channel with RabbitMQ cleanly by issuing the
		Channel.Close RPC command.

		"""
		log.debug('Closing the channel')
		self._channel.close()

	def run(self):
		"""Run the example consumer by connecting to RabbitMQ and then
		starting the IOLoop to block and allow the SelectConnection to operate.

		"""
		self._connection = self.connect()
		self._connection.ioloop.start()

	def stop(self):
		"""Cleanly shutdown the connection to RabbitMQ by stopping the consumer
		with RabbitMQ. When RabbitMQ confirms the cancellation, on_cancelok
		will be invoked by pika, which will then closing the channel and
		connection. The IOLoop is started again because this method is invoked
		when CTRL-C is pressed raising a KeyboardInterrupt exception. This
		exception stops the IOLoop which needs to be running for pika to
		communicate with RabbitMQ. All of the commands issued prior to starting
		the IOLoop will be buffered but not processed.

		"""
		log.debug('Stopping')
		self._closing = True
		self.stop_consuming()
		self._connection.ioloop.start()
		log.debug('Stopped')

	def close_connection(self):
		"""This method closes the connection to RabbitMQ."""
		log.debug('Closing connection')
		self._connection.close()



"""
	SeaBird CTD is meant to handle reading data from a stationary CTD logger. Many packages exist for profile loggers,
	but no options structured for use on a stationary logger or reading. This package is designed to remain directly connected
	to a CTD while it's logging and can be extended to support communication with different versions of the Seabird CTD
	loggers with different capabilities and command descriptions.
"""

__author__ = "nickrsan"
from seabird_ctd.version import version as __version__

import time
from datetime import timezone
import pytz
import datetime
import os
import re
import logging
import traceback

import six

import serial

#logging.basicConfig(level=logging.DEBUG)

try:
	import pika
except ImportError:
	interrupt = None
	logging.debug("pika not available, can't use RabbitMQ to check for messages")

try:
	from seabird_ctd import interrupt
	from multiprocessing import Process
except:
	interrupt = None
	logging.warning("Unable to load a required module for interrupt queue. Can be ignored if not using RabbitMQ, but this message is unusual regardless.")

class SBE37S(object):
	"""
		Handles the SBE37S commands
	"""
	def __init__(self, main_ctd):
		self.max_samples = 3655394  # needs verification. This is just a guess based on SBE39
		self.keys = ("pressure", "conductivity", "temperature", "salinity", "datetime")
		self.ctd = main_ctd

	def set_datetime(self):
		dt = datetime.datetime.now(timezone.utc)
		return ["DATETIME={}".format(dt.strftime("%m%d%Y%H%M%S"))]

	def sample_interval(self, interval):
		return "SAMPLEINTERVAL={}".format(interval)

	def retrieve_samples(self, start, end):
		return [
			"OUTPUTFORMAT=1",
			"GetSamples:{},{}".format(start,end)
		]

	def parse_status(self, status_message):
		return_dict = {}
		return_dict["full_model"] = status_message[1].split(" ")[0]
		return_dict["serial_number"] = status_message[1].split(" ")[4]

		voltages = re.match("vMain\s+=\s+(\d+\.\d+),\s+vLith\s+=\s+(\d+\.\d+)", status_message[2])
		return_dict["battery_voltage"] = voltages.group(0)
		return_dict["lithium_voltage"] = voltages.group(1)

		return_dict["sample_number"] = status_message[3].split(", ")[0].split(" = ")[1].replace(" ", "")
		return_dict["is_sampling"] = True if status_message[4] == "logging data" else False
		return_dict["salinity_output"] = True if "output salinity" in status_message else False
		return return_dict

	def record_regex(self):
		"""
			Handles generation of the regex for the data records. If salinity output is turned on, handles that correctly.
			STILL TO DO - handle other optional values, such as sound velocity - what happens if sound velocity is on,
			but salinity is off, or if both are on?
		:return:
		"""
		if self.ctd.salinity_output:
			salinity_insert = "\s+(?P<salinity>?\d+\.\d+),"
		else:
			salinity_insert = ""

		self.regex = "(?P<temperature>-?\d+\.\d+),\s+(?P<pressure>-?\d+\.\d+),"+salinity_insert+"\s+(?P<datetime>\d+\s\w+\s\d{4},\s\d{2}:\d{2}:\d{2})"
		return self.regex

class SBE39(object):
	def __init__(self, main_ctd):
		self.max_samples = 3655394
		self.keys = ("temperature", "pressure", "datetime")
		self.ctd = main_ctd

	def set_datetime(self):
		dt = datetime.datetime.now(timezone.utc)
		return ["MMDDYY={}".format(dt.strftime("%m%d%y")), "HHMMSS={}".format(dt.strftime("%H%M%S"))]

	def sample_interval(self, interval):
		return "INTERVAL={}".format(interval)

	def retrieve_samples(self, start, end):
		return [
			"DD{},{}".format(start, end)
		]

	def record_regex(self):
		self.regex = "(?P<temperature>-?\d+\.\d+),\s+(?P<pressure>-?\d+\.\d+),\s+(?P<datetime>\d+\s\w+\s\d{4},\s\d{2}:\d{2}:\d{2})"
		return self.regex

	def parse_status(self, status_message):
		return_dict = {}
		return_dict["full_model"] = status_message[1].split("   ")[0]
		return_dict["serial_number"] = status_message[1].split("   ")[1]
		return_dict["battery_voltage"] = status_message[2].split(" = ")[1]
		return_dict["sample_number"] = status_message[5].split(", ")[0].split(" = ")[1]
		return_dict["is_sampling"] = True if status_message[3] == "logging data" else False
		return return_dict

supported_ctds = {
	"SBE37S": "SBE37S",
	"SBE 39": "SBE39"
}  # name the instrument will report, then class name. could also do this with 2-tuples.

class CTDConfigurationError(BaseException):
	pass

class CTDOperationError(BaseException):
	pass

class CTDConnectionError(BaseException):
	pass

class CTDUnsupportedError(BaseException):
	pass


class CTD(object):
	def __init__(self, COM_port=None, baud=9600, timeout=5, setup_delay=2):
		"""
		If COM_port is not provided, checks for an environment variable named SEABIRD_CTD_PORT. Otherwise raises
		CTDConnectionError

		"""

		if COM_port is None:
			if "SEABIRD_CTD_PORT" in os.environ:
				COM_port = os.environ["SEABIRD_CTD_PORT"]
			else:
				raise CTDConnectionError("SEABIRD_CTD_PORT environment variable is not defined. Don't know what COM port"
										 "to connect to for CTD data. Can't collect CTD data.")

		self.log = logging.getLogger("seabird_ctd.{}".format(COM_port))  # lets you tune into all seabird_ctd logging, or just this one by com port

		self.last_sample = None
		self.is_sampling = None  # starts at None, will be set when DS is run.
		self.sample_number = None  # will be set by DS command
		self.com_port = COM_port

		self.ctd = serial.Serial(COM_port, baud, timeout=timeout)
		time.sleep(setup_delay)  # give it time to init

		self.command_object = None
		self.handler = None  # will be set if self.listen is called
		self.held_records = []

		self.model = None  # to be set later

		# STATUS INFO TO BE SET BY .status()
		self.full_model = None
		self.serial_number = None
		self._battery_voltage = None  # actual battery voltage is function so we can keep track of changes
		self._original_voltage = None
		self.lithium_voltage = None
		self.sample_number = None
		self.is_sampling = None
		self.salinity_output = None

		# INTERNAL FLAGS AND INFO
		self.rabbitmq_server = None
		self._stop_monitoring = False  # a flag that will be monitored to determine if we should stop checking for commands
		self._close_connection = False  # same as previous

		self.last_status = None  # will be set each time status is run
		self.determine_ctd_model()
		if self.last_status is None:  # we check it here because determine_ctd_model might run it if it needs to. Don't waste time doubling up.
			self.status()  # will fill some fields in so we know what sample it's on, etc

	@property
	def battery_voltage(self):
		return self._battery_voltage

	@battery_voltage.setter
	def battery_voltage(self, new_value):
		if self._battery_voltage is None:
			self._original_voltage = float(new_value)

		self._battery_voltage = float(new_value)

	@property
	def battery_change(self):
		return (self._battery_voltage - self._original_voltage) / self._original_voltage

	def determine_ctd_model(self):
		try:
			self.log.debug("Waking CTD")
			self.wake()
			ctd_info = self._read_all() # TODO: Is the data returned from wake always junk? What if this is starting up after setting the CTD to autosample, then the server crashes and is reconnecting
		except UnicodeDecodeError:
			raise CTDConfigurationError("Unable to decode response from CTD. Try a different baud setting and ensure it matches the CTD's internal configuration")

		self.log.debug("CTD responded with {}".format(ctd_info))

		for line in ctd_info:  # it should write out the model when you connect - this works unless it hasn't timed out since last connection
			if line in supported_ctds.keys():
				self.log.info("Model Found: {}".format(line))
				self.command_object = globals()[supported_ctds[line]]()  # get the object that has the command info for this CTD
				self.model = line

		if not self.command_object:  # if we didn't get it from startup, issue a DS to determine it and peel it off the first part
			self.log.debug("Trying backup method to determine model")
			ds = self.send_command("DS")
			self.log.debug("CTD responded with {}".format(ds))
			self.model = ds[1][:6]  # it'll be the first 6 characters of the
			self.log.debug("Best guess for model is {}".format(self.model))

			if self.model not in supported_ctds:
				raise CTDUnsupportedError("Model '{}' is not supported. If this doesn't match the model you have, then the model information did not correctly parse. You may try again.".format(self.model))

			self.command_object = globals()[supported_ctds[self.model]]()

		if not self.command_object:
			raise CTDConnectionError("Unable to wake CTD or determine its type. There could be a connection error or this is currently plugged into an unsupported model")

	def send_command(self, command=None, length_to_read="ALL"):
		if command:
			self.ctd.write(six.b('{}\r\n'.format(command)))  # doesn't seem to work unless we pass it a windows line ending. Sends command, but no results
			time.sleep(1)  # was this actually necessary? Try removing it.

		self.log.debug("{} bytes in waiting".format(self.ctd.in_waiting))
		if self.ctd.in_waiting > 0:  # if the CTD sent data and we haven't read it yet
			# reads after sending by default so that we can determine if there was a timeout
			if length_to_read == "ALL":
				data = b""
				while self.ctd.in_waiting > 0:
					new_data = self.ctd.read(self.ctd.in_waiting)  # if we're expecting quite a lot, then keep reading until we get nothing
					data += new_data
			elif length_to_read is None:
				return  # for some commands, such as QS, we want to not attempt a read after sending
			else:
				data = self.ctd.read(length_to_read)

			response = self._clean(data)

			if self.is_sampling:  # any time we read data, if we're sampling, we should check it for records
				self.check_data_for_records(response)

			if "timeout" in response or self.model in response:  # if we got a timeout the first time, then we should be reconnected. Rerun this function and return the results
				# we can safely check for the model because by the time it's all a list, no single line in the list is *only* the model unless we received a timeout
				self.log.debug("Received unexpected response. Resending command (if any - input was {}). Response was {}".format(command, response))
				return self.send_command(command, length_to_read)
			else:
				return response

	def _read_all(self):
		return self.send_command(command=None, length_to_read="ALL")

	def _clean(self, response):
		return response.decode("utf-8").split("\r\n")  # first element of list should now be the command, but we'll let the caller filter that

	def set_datetime(self, raise_error=False):
		"""
			Attempt to set the datetime. Logs a warning if that can't be done because the sampler is logging. If you want
			it to raise an exception

		:param raise_error: When False, the default, this function warns if it can't set the datetime. When True, raises CTDOperationError
		:return:
		"""

		if self.is_sampling:
			msg = "Can't change datetime while CTD is sampling. Issue a stop command (or call ctd.stop_autosample()), run set_datetime again, and restart sampling in order to change datetime."
			if raise_error:
				raise CTDOperationError(msg)
			else:
				self.log.warning(msg)
				return

		datetime_commands = self.command_object.set_datetime()
		for command in datetime_commands:
			self.log.info("Setting datetime: {}".format(command))
			self.send_command(command, length_to_read=None)

	def take_sample(self):
		self.last_sample = self.send_command("TS", length_to_read=1000)
		return self.last_sample

	def _filter_samples_to_data(self,):
		pass

	def sleep(self):
		self.send_command("QS", length_to_read=None)

	def wake(self):
		self.send_command("\r\n", length_to_read=None)  # Send a single character to wake the device, get the response so that we clear the buffer

	def status(self):
		try:
			if self.ctd.in_waiting > 0:  # any current waiting characters will make parsing weird.
				self._read_all()  # clear the input buffer, check for any data in the pipeline

			status = self.send_command("DS")
			status_parts = self.command_object.parse_status(status)
			for key in status_parts:  # the command object parses the status message for the specific model. Returns a dict that we'll set as values on the object here
				setattr(self, key, status_parts[key])  # set each returned value as an attribute on this object

			self.last_status = datetime.datetime.now(timezone.utc)

			self.log.info(status)
		except:  # no matter what exception is raised, we'll most likely want to roll through this
			if self.last_status is not None:  # if this isn't our first check of the status, then warn, but keep going
				self.log.warning("Failed to retrieve or parse status message. Proceeding, but some information, such as battery voltage, may be out of date. CTD response was {}. Error given was {}".format(status, traceback.format_exc()))
			else:  # if it is our first time getting status data, raise the exception up, because we shouldn't proceed without it
				raise

		return status

	def setup_interrupt(self, rabbitmq_server, username, password, vhost, queue=None):
		"""
			Used to set the monitoring functions to use the interrupt method, which allows messages to be passed to the CTD
			from the user even while monitoring for data. By default, if this function is not called, then the monitoring code
			cannot be interrupted and simply runs until the user cancels it.

		:param rabbitmq_server:
		:param username:
		:param password:
		:param vhost:
		:param queue:
		:return:
		"""

		if not interrupt:
			raise CTDConfigurationError("Can't set up interrupt unless Pika Python package is installed. Pika was not found")

		if queue is None:
			queue = self.com_port

		self.rabbitmq_server = rabbitmq_server
		self.interrupt_connection = interrupt.RabbitConsumer(self.rabbitmq_server)
		self.interrupt_connection.username = username
		self.interrupt_connection.password = password
		self.interrupt_connection.vhost = vhost
		self.interrupt_connection.QUEUE = queue

	def check_interrupt(self):
		"""
			Should return True if we're supposed to stop listening or False if we shouldn't

		:return:
		"""

		if not interrupt or not self.rabbitmq_server:  # if pika isn't available, we can't check for messages, so there's always no interrupt
			return False  # these are silently returned False because it means that they didn't try to configure the interrupt

		if self._stop_monitoring:
			return True

		return False  # default is no interrupt

	def start_autosample(self, interval=60, realtime="Y", handler=None, no_stop=False):
		"""
			This should set the sampling interval, then turn on autosampling, then just keep reading the line every interval.
			Before reading the line, it should also check for new commands in a command queue, so it can see if it's
			should be doing something else instead.

			We should do this as an event loop, where we create a celery task to check the line. That way, control can flow
			from here and django can do other work in the meantime. Otherwise, we can have a separate script that does
			the standalone django setup so it can access the models and the DB, or we can just do our own inserts since
			it's relatively simple code here.

		:param interval:  How long, in seconds, should the CTD wait between samples
		:param realtime: Two possible values "Y" and "N" indicating whether the CTD should return results as soon as it collects them
		:param handler: This should be a Python function (the actual object, not the name) that takes a list of dicts as its input.
		 				Each dict represents a sample and has keys for "temperature", "pressure", "conductivity",
		 				and "datetime", as appropriate for the CTD model. It'll skip parameters the CTD doesn't collect.
		 				The handler function will be called whenever new results are available and can do things like database input, etc.
		 				If realtime == "Y" then you must provide a handler function.
		:param no_stop: Allows you to tell it to ignore settings if it's already sampling. If no_stop is True, will just
						start listening to new records coming in (if realtime == "Y"). That way, the CTD won't stop sampling
						for any length of time, but will retain prior settings (ignoring the new interval).
		:return:
		"""

		self.log.info("Initiating autosampling")

		if self.is_sampling and not no_stop:
			self.stop_autosample()  # stop it so we can set the parameters - stop_autosample sets the flag

		if not self.is_sampling:  # will be updated if we successfully stop sampling
			self.send_command(self.command_object.sample_interval(interval), length_to_read=None)  # set the interval to sample at
			self.send_command("TXREALTIME={}".format(realtime), length_to_read=None)  # set the interval to sample at
			self.send_command("STARTNOW", length_to_read=100)  # start sampling - we need to read to clear the buffer
			self.send_command("\r\n", length_to_read=None)  # send a newline so that we get a new prompt again
			self.status()  # make sure the status data is up to date. Doing it this way ratehr than setting manually so that if it failed for some reason, the object would still be correct
							# if is_sampling doesn't get updated here, data reading won't work correctly, so this is important

		if realtime == "Y":
			if not handler:  # if they specified realtime data transmission, but didn't provide a handler, abort.
				raise CTDConfigurationError("When transmitting data in realtime, you must provide a handler function that accepts a list of dicts as its argument. See documentation for start_autosample.")
			else:
				self.handler = handler  # make it available to anything else that finds records. Whenever we run any command we need to check for records because it's possible we'll get the output before something else

			self.listen(interval)

	def listen(self, interval=60):
		"""
			Continuously polls for new data on the line at interval. Used when autosampling with realtime transmission to
			receive records.

			See documentation for start_autosample for full documentation of these parameters. This function remains
			part of the public API so you can listen on existing sampling if the autosampler is already configured.

			:param handler: (See start_autosample documentation). A functiion to process new records as they are available.
			:param interval: The sampling interval, in seconds.

			:return:
		"""

		self.log.info("Starting listening loop for CTD data")

		self.check_interval = interval

		def interrupt_handler(ch, method, properties, body):
			"""
				This function is done as a closure because we want it to be able to access attributes of the class instance
				namely, the open pipe to the serial port, as well as knowing which COM port we're on so that it can load
				a RabbitMQ queue with the same name.

				For stopping and closing the connection, this function sets flags on the object that are then responded
				to next time the listen loop wakes up to check the CTD. Flags will be handled before reading the CTD.
				Commands sent to the device will be sent immediately.

				Commands that can be sent to this program are READ_DATA, STOP_MONITORING and DISCONNECT. STOP_MONITORING will just
				stop listening to the CTD, but will leave the CTD in its current autosample configuration.
				If you want stop the CTD from logging, send a Stop command that's appropriate for the CTD model you're
				using (usually STOP) then send a STOP command through the messaging to this program so it will stop checking
				the CTD for new data. DISCONNECT is equivalent to STOP_MONITORING, but also puts the CTD to sleep and
				closes the connection to the CTD.

			:param ch:
			:param method:
			:param properties:
			:param body:
			:return:
			"""

			logging.info(" [x] Received Command via Queue %r" % body)

			body = body.decode("utf-8")
			if body == "READ_DATA":
				if not self._stop_monitoring:
					self.read_records()
			elif body == "DS" or body == "STATUS":
				status = self.status()
				self.log.info(status)
			elif body == "STOP_MONITORING":  # just stop monitoring - once monitoring is stopped, they can connect to the device to stop autosampling
				self.log.info("Stopping monitoring of records")
				self._stop_monitoring = True
				self.interrupt_connection.stop()
			elif body == "DISCONNECT":
				self.log.info("Shutting down monitoring script and closing connection to CTD")
				self.interrupt_connection.stop()
				self.close()  # puts CTD to sleep and closes connection
			else:
				data = self.send_command(body)
				self.log.info(data)

			self.interrupt_connection.acknowledge_message(method.delivery_tag)

		if interrupt and self.rabbitmq_server:
			self.log.debug("Starting interrupt listening loop")

			self.interrupt_connection.handler = interrupt_handler

			self._stop_monitoring = False  # reset the flags that are used to determine if we should stop
			self._close_connection = False

			self.log.debug("Spinning up CTD read loop process")
			self.p = Process(target=interrupt_checker, kwargs={"server": self.rabbitmq_server,
							  "username": self.interrupt_connection.username,
							  "password": self.interrupt_connection.password,
							  "vhost": self.interrupt_connection.vhost,
							  "queue": self.interrupt_connection.QUEUE,
							  "interval": self.check_interval})
			self.p.start()
			self.log.info(self.p.join(5))  # give it 5 seconds of blocking to detect if the process exits

			self.interrupt_connection.run()  # starts listening for commands

		else:  # if no interrupt loop
			while self.check_interrupt() is False:
				self.read_records()
				time.sleep(interval)

	def read_records(self):

		self.log.debug("Checking for CTD data")

		data = self._read_all()  # this will automatically check for data

		if (datetime.datetime.now(timezone.utc) - self.last_status).total_seconds() > 3600:  # if it's been more than an hour since we checked the status, refresh it so we get new battery stats
			self.status()  # this can be run while the device is logging

	def check_data_for_records(self, data):
		records = self.find_and_insert_records(data)

		if not self.handler:  # if it's sampling and we've received records, but not yet configured a handler, hold onto the records until a handler is configured
			self.held_records += records
		else:
			if len(self.held_records) > 0:
				self.handler(self.held_records)  # handle the held records first, then zero out the list
				self.held_records = []
			self.handler(records)  # Call the provided handler function to do whatever the caller wants to do with the data

	def find_and_insert_records(self, data):
		"""
			Given the parsing regex defined in the command object for this CTD, runs it and returns a list of dicts
			containing sampled values. The dict will always have the keys defined in the command object's "keys" attribute.
			This varies model to model, so your code may need to check for keys, depending. In the event of values that
			are not always enabled (like salinity, or sound velocity), the keys are still present, but set to None if
			those values are not enabled.
		:param data:
		:return:
		"""
		records = []
		for element in data:
			matches = re.search(self.command_object.record_regex(), element)
			if matches is None:
				continue

			new_model = {}
			for key in self.command_object.keys:
				try:
					value = matches.group(key)
					if key == "datetime":
						dt_object = datetime.datetime.strptime(str(value), "%d %b %Y, %H:%M:%S")
						value = pytz.utc.localize(dt_object)
					new_model[key] = value
				except IndexError:
					new_model[key] = None  # it just means that item isn't supported or enabled on this sample

			records.append(new_model)

		return records

	def stop_autosample(self):
		self.log.info("Stopping existing autosampling")
		self.send_command("STOP", length_to_read=None)
		self.status()

		if self.is_sampling is not False:
			raise CTDOperationError("Unable to stop autosampling. If you are starting autosampling, parameters may not be updated")

	def catch_up(self):
		"""
			Meant to pull in any records that are missing on startup of this script - if the autosampler runs in the
			background, then while the script is offline, new data is being stored in flash on the device. This should
			pull in those records.

			NOTE - this command STOPS autosampling if it's running - it must be restarted on its own.

		:return:
		"""

		self.stop_autosample()

		commands = self.command_object.retrieve_samples(0, self.sample_number)
		for command in commands:
			results = self.send_command(command)  # if we have multiple commands, only the later one will have data

		# we now have all the samples and can parse them so that we can insert them.

	def close(self):
		"""
			Put the CTD to sleep and close the connection

		:return:
		"""

		self.sleep()
		self.ctd.close()
		self.ctd = None

	def __del__(self):
		""""
			Check if the CTD is still open. If the user called close on their own, then this will create an error.

		"""
		if self.ctd:
			self.close()

def interrupt_checker(server, username, password, vhost, queue, interval):
	"""
		When using the interrupt method, this code handles the scheduling of the actual checking by connecting
		to RabbitMQ and sending READ commands every interval

	:param server:
	:param username:
	:param password:
	:param vhost:
	:param queue:
	:param interval:
	:return:
	"""

	connection = pika.BlockingConnection(pika.ConnectionParameters(host=server, virtual_host=vhost,
																   credentials=pika.PlainCredentials(username, password)))
	channel = connection.channel()

	channel.queue_declare(queue=queue)

	command = "READ_DATA"
	while True:
		channel.basic_publish(exchange='seabird',
						  routing_key='seabird',
						  body=command)

		time.sleep(interval)


import unittest

import seabird_ctd


class CTDTest(unittest.TestCase):
	def setUp(self):
		self.ctd = seabird_ctd.CTD(baud=4800)

	def test_data_read(self):

		try:
			self.assertTrue(hasattr(self.ctd, "sample_number"))  # this gets assigned after initialization, and comes in when status info is read
		finally:
			self.ctd.sleep()
			self.ctd.ctd.close()

	def test_ctd_deletion(self):
		"""
			We had an issue where upon exiting python or deleting the CTD object, it raised an Attribute Error because
			it attempted to read from the CTD after sending the QS command. Resolved by setting a parameter on send_command
			so that it can send without reading in those situations. This test makes sure deletion doesn't raise an exception.

			What's unfortunate is that due to the way tthe exception is raised (during a __del__ call), it doesn't actually
			propagate up. So, this test will always pass, but is here so that should an exception occur during object
			deletion, it points back to this test.
		:return:
		"""
		try:
			del self.ctd
		except TypeError:
			self.assertFalse(True)  # raised AttributeError on deleting CTD object
		except:
			self.assertFalse(True)  # raised some exception on deleting CTD object

		self.assertFalse(False)  # does not raise an exception upon deleting CTD object

	def test_data_reconnect(self):
		# we should run a status command, then wait 3 minutes for it to go to sleep, then run it again and make sure we
		# get the results of the command.
		pass


if __name__ == "__main__":
	unittest.main()


"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

#from seabird_ctd import __version__ as version  # we need to move this to a place where it doesn't load dependencies before we can use this

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="seabird_ctd",
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version="0.2.2",

    description='Tools for communicating with, retrieving data from, and managing a fixed Seabird CTD deployment',
    long_description="""
    The Seabird CTD library communicates with, controls, and reads data directly off of a Seabird CTD. It has been designed for the legacy SBE 39 and the modern SBE 37+, but can be extended to support other models easily. Other CTD packages focus on working with the data, especially profiles, once it's downloaded, but this package is designed to allow for direct control and monitoring during a longterm fixed deployment. The authors use this code to feed the CTD data into a database via django in realtime and implement the monitoring as a management command.
    
    In-development documentation can be found at https://seabird-ctd.readthedocs.io/en/latest/
    
    Installation documentation can be found at https://bitbucket.org/b195m/seabird_ctd/ - note that for installation,
    the package name is seabird-ctd (eg: `pip install seabird-ctd`), but when using, the package name is seabird_ctd.
    
    This code was developed by Nick Santos for The University of Oregon with funding from the National Science Foundation.
    """,

    # The project's main homepage.
    url='https://bitbucket.org/b195m/seabird_ctd/',

    # Author details
    author='Nick Santos',
    author_email='nick@nicksantos.com',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    # What does your project relate to?
    keywords='seabird ctd serial temperature pressure conductivity',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['six', 'pyserial', 'pika'],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    #extras_require={
    #    'dev': ['check-manifest'],
    #    'test': ['coverage'],
    #},

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    #package_data={
    #    'sample': ['package_data.dat'],
    #},

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    #data_files=[('my_data', ['data/data_file'])],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    #entry_points={
    #    'console_scripts': [
    #        'sample=sample:main',
    #    ],
    #},
)